from ..config import ITERATION_FORMATTING
from .tokenize import Tokenizer
from .errors import (
    MergeError,
    IterationPaddingMisMatch,
    IterationSetParseError,
    StringReprFormatError,
)


class IterationRange(object):
    def __init__(self, first, last):
        self.first = first
        self.last = last

    def __str__(self):
        return "{first}-{last}".format(first=self.first, last=self.last)

    def __repr__(self):
        return str(self)


class IterationSet(object):
    def __init__(self, *iterations, step=1, padding=0):
        self._iteration_range_groups = {}
        self.step = step
        self.padding = padding

        for iteration in iterations:
            self.add(iteration)

    def __str__(self):
        joined_ranges = self.str_ranges()

        if self.is_single():
            return joined_ranges
        else:
            return "{{{}}}".format(joined_ranges)

    def __repr__(self):
        return str(self)

    def __iter__(self):
        for iteration_range in self.iteration_ranges():
            for iteration_int in range(
                iteration_range.first, iteration_range.last + 1, self.step
            ):
                yield iteration_int

    def str_iter(self):
        for item in self:
            yield str(item).zfill(self.padding)

    def str_ranges(self):
        string_ranges = []
        for iteration_range in self.iteration_ranges():
            if self.is_single():
                iteration = str(iteration_range.first).zfill(self.padding)
                string_ranges.append("{iteration}".format(iteration=iteration))
            else:
                first = str(iteration_range.first).zfill(self.padding)
                last = str(iteration_range.last).zfill(self.padding)
                if first == last:
                    string_ranges.append("{first}".format(first=first))
                else:
                    string_ranges.append(
                        "{first}-{last}".format(first=first, last=last)
                    )

        return ",".join(string_ranges)

    def __len__(self):
        length = 0
        for group in self.iteration_ranges():
            group_length = group.last - group.first + 1
            length += group_length
        return length

    def __eq__(self, other):

        if not isinstance(other, IterationSet):
            return False

        # Check if the groups are the same length
        if len(self) != len(other):
            return False

        # Iterate over the iteration range groups and make sure they are equal
        for group1, group2 in zip(self.iteration_ranges(), other.iteration_ranges()):
            if group1.first != group2.first or group1.last != group2.last:
                return False

        return True

    @staticmethod
    def build(range_groups, padding=0):

        # TODO: protect against bad groups
        iteration_set = IterationSet(padding=padding)

        for range_group in range_groups:
            first = range_group[0]
            last = range_group[-1]

            if first is None or last is None:
                raise ValueError(
                    "When building an IterationSet, each frame group must "
                    "have a value first and last value. "
                    "Recieved: first: {}, last: {}".format(first, last)
                )

            if not isinstance(first, int) or not isinstance(last, int):
                raise TypeError(
                    "When building an IterationSet, "
                    "each first and last item of a frame group "
                    "must be ints"
                )

            if not first <= last:
                raise ValueError(
                    "When building an IterationSet, "
                    "first int of a frame group must be "
                    "less than or equal to the last"
                )

            iteration_set._iteration_range_groups[first] = IterationRange(first, last)

        return iteration_set

    @staticmethod
    def from_string(input_string, token_string="{range_groups:[\d\-\,]+}"):

        parsed = Tokenizer.tokenize(input_string, [token_string])

        # Parse padding
        padding = int(parsed.get("padding", 0))

        # Handle when first and last are parsed tokens
        first = parsed.get("first")
        last = parsed.get("last")
        if first or last:
            if not first or not last:
                raise IterationSetParseError(
                    "First and last tokens must be used together. "
                    "Parsed tokens: {}".format(parsed.keys())
                )

            if not padding:
                if len(first) != len(last):
                    raise IterationPaddingMisMatch(
                        "Parsed first frame padding != parsed last frame padding. "
                        "Recieved: first_frame: {}, padding: {}"
                        "last_frame: {} padding: {}".format(
                            first, len(first), last, len(last)
                        )
                    )
                padding = len(first)

            first = int(first)
            last = int(last)
            if not first <= last:
                raise IterationSetParseError(
                    "First must be less than last. "
                    "Recieved: first: {}, last: {}".format(first, last)
                )

        # If first/last isn't included
        else:
            first = 0
            last = 0

        # Handle range groups
        range_groups_repr = parsed.get("range_groups")

        range_groups = []
        if range_groups_repr:
            # expect a string like: 1004-1007,1009,1011-1020
            groups = range_groups_repr.split(",")
            for group in groups:
                range_groups.append(tuple(int(x) for x in group.split("-")))
        else:
            range_groups.append((first, last))

        # If there isn't a first, last or range group token, then this will be empty
        # We need range groups to build a useable IterationSet
        if not range_groups:
            raise IterationSetParseError(
                "Unable to parse iteration set information such as: "
                "first, last, or even range groups"
            )

        return IterationSet.build(range_groups, padding=padding)

    @property
    def first(self):
        iteration_range_start_points = self.iteration_range_group_keys()

        if not iteration_range_start_points:
            return 0

        return iteration_range_start_points[0]

    @property
    def last(self):
        iteration_range_start_points = self.iteration_range_group_keys()

        if not iteration_range_start_points:
            return 0

        return self.get_iteration_range_group(iteration_range_start_points[-1]).last

    @staticmethod
    def merge(iteration_set_1, iteration_set_2):

        if iteration_set_1.padding != iteration_set_2.padding:
            raise MergeError(
                "IterationSet1 (padding: {padding1}) and IterationSet2 (padding: {padding2}) do not match".format(
                    padding1=iteration_set_1.padding, padding2=iteration_set_2.padding
                )
            )

        resulting_iteration_set = IterationSet(padding=iteration_set_1.padding)
        resulting_iteration_set.add(*iteration_set_1)
        resulting_iteration_set.add(*iteration_set_2)

        return resulting_iteration_set

    def token_data(self):
        return dict(
            first=str(self.first).zfill(self.padding),
            last=str(self.last).zfill(self.padding),
            unpadded_first=self.first,
            unpadded_last=self.last,
            padding=self.padding,
            range_groups=self.str_ranges(),
        )

    def is_single(self):

        iteration_ranges = list(self.iteration_ranges())
        if len(iteration_ranges) > 1 or len(iteration_ranges) <= 0:
            return False

        iteration_range = iteration_ranges[0]
        return iteration_range.first == iteration_range.last

    def is_empty(self):
        return self.is_single() and self.first == 0 and self.last == 0

    def format(self, iteration_format_key):

        if not iteration_format_key or self.is_single():
            return str(self)

        iteration_format = ITERATION_FORMATTING.get(iteration_format_key)
        if not iteration_format:
            raise ValueError(
                f"{iteration_format_key} is not a valid IterationFormatter key. "
                f"Valid keys are: {', '.join(ITERATION_FORMATTING.keys())}"
            )

        try:
            return Tokenizer.format(iteration_format, self.token_data())
        except ValueError as e:
            raise StringReprFormatError(e)

    def sanitize_item(self, item):
        """
        IterationSets can accept ints or strings representing ints.
         This method is used to ensure that.
        """

        # Supporting only adding string digits to iteration sets for now
        if not isinstance(item, (str, int)):
            raise TypeError(
                "Expecting padded digits or ints -- recieved: {}".format(type(item))
            )

        # Set up padding if not set up (new iteration set init)
        if isinstance(item, str) and not item.isdigit():
            raise ValueError(
                "The item is a string but, not a digit -- received: {} -- {}".format(
                    item, type(item)
                )
            )

        if isinstance(item, str) and self.padding and self.padding != len(item):
            raise IterationPaddingMisMatch(
                'Iteration "{iteration}" and this IterationSet have a padding mis-match -- '
                "IterationSet = {iteration_set_padding}, Iteration = {iteration_padding}".format(
                    iteration=item,
                    iteration_set_padding=self.padding,
                    iteration_padding=len(item),
                )
            )

        return int(item)

    def add(self, *iterations):

        for iteration in iterations:

            # Post validation adding a iteration
            iteration_int = self.sanitize_item(iteration)

            # Set up padding if not set up (new iteration set init)
            if isinstance(iteration, str) and not self.padding:
                self.padding = len(iteration)

            if not self.get_iteration_range_group(iteration_int):
                self._iteration_range_groups[iteration_int] = IterationRange(
                    first=iteration_int, last=iteration_int
                )

            # Consolidate IterationRange groups
            previous_index = None
            for current_index in self.iteration_range_group_keys():

                move_previous_pointer = True
                if isinstance(previous_index, int):

                    previous_group = self.get_iteration_range_group(previous_index)
                    current_group = self.get_iteration_range_group(current_index)

                    if previous_group and current_group:
                        # If the current group's first iteration is in the range of the previous iteration range, consolodate them
                        if (
                            previous_group.last + 1 >= current_group.first
                            and current_group.first >= previous_group.first
                        ):

                            # Whichever group has the largest last iteration will become the last iteration of the consolodated group
                            consolodated_last = (
                                previous_group.last
                                if previous_group.last > current_group.last
                                else current_group.last
                            )

                            previous_group.last = consolodated_last

                            del self._iteration_range_groups[current_index]
                            move_previous_pointer = False

                if move_previous_pointer:
                    previous_index = current_index

    def iteration_range_group_keys(self):
        # Sorting by keys to ensure the correct iteration group ordering (all keys should be ints)
        return sorted(self._iteration_range_groups.keys())

    def iteration_ranges(self):
        for start_iteration in self.iteration_range_group_keys():
            yield self._iteration_range_groups.get(start_iteration)

    def get_iteration_range_group(self, index):
        return self._iteration_range_groups.get(index)

    def has(self, item):
        """
        Iterate over the range groups and see if the item exists in any of them
        """

        try:
            item = self.sanitize_item(item)
        except (ValueError, TypeError, IterationPaddingMisMatch):
            return False

        for range_group in self.iteration_ranges():
            if item >= range_group.first and item <= range_group.last:
                return True
        return False



import os
import re
from glob import glob
from sre_constants import error as RegexError
from collections import defaultdict, OrderedDict
from itertools import chain, zip_longest
from .tokenize import Tokenizer
from .errors import MergeError, FilesNotLocal
from .iteration import IterationSet


class PartitionedFilepath(object):
    """A representation of a filepath that's split into interation sets and strings.

    The PartitionedFilepath class takes a filepath string and
     splits it into sections of strings and sections of digits.

    The digits are stored in IterationSets so when merged, the
     iteration sets can be combined.

    Args:
        filepath (str): '/path/to/dir/or/file.txt'
    """

    def __init__(self, filepath_string):

        self._strings = OrderedDict()
        self._iteration_sets = OrderedDict()

        self._setup(filepath_string)

    def _setup(self, filepath_string):

        num_string = ""
        alpha_string = ""

        counter = 0
        for item in filepath_string:

            if isinstance(item, str):

                if item.isdigit():
                    if alpha_string:
                        self._strings[counter] = alpha_string
                        counter += 1
                        alpha_string = ""

                    num_string += item

                else:
                    if num_string:
                        iteration_set = IterationSet(num_string)
                        self._iteration_sets[counter] = iteration_set
                        counter += 1
                        num_string = ""

                    alpha_string += item

        # If there were any partitions building before the loop ended
        if num_string:
            iteration_set = IterationSet(num_string)
            self._iteration_sets[counter] = iteration_set
            counter += 1

        if alpha_string:
            self._strings[counter] = alpha_string
            counter += 1

    def __str__(self):
        return "".join([str(item) for item in self])

    def __repr__(self):
        return str(self)

    def __len__(self):
        return len(self._strings) + len(self._iteration_sets)

    def __iter__(self):
        """
        The strings and iteration sets are kept in different DefaultDicts,
         iterating over a partitioned path reassembles the 2 DefaultDicts and,
        yields them.
        """
        merged_items = self._strings.copy()
        merged_items.update(self._iteration_sets)

        for i in sorted(merged_items.keys()):
            yield merged_items[i]

    def items(self):
        for item in self:
            yield item

    @staticmethod
    def from_items(items):
        new_partitioned_path = PartitionedFilepath("")
        for i, item in enumerate(items):
            if isinstance(item, IterationSet):
                new_partitioned_path._iteration_sets[i] = item
            else:
                new_partitioned_path._strings[i] = item

        return new_partitioned_path

    @staticmethod
    def from_str_repr(formatted_string, iteration_format_string):

        items = PartitionedFilepath.split(formatted_string, iteration_format_string)
        return PartitionedFilepath.from_items(items)

    @staticmethod
    def split(input_string, format_string):
        """Splits a string by iteration sets.

        This method expects strings of a certain format.

        It will parse the string for certain template patterns,
         convert those into regex and split the string via re.split

        Args:
            formatted_string (str): A string to be split

        Returns:
            list: list of strings representing IterationSets
            list: list of strings
        """

        # Parses the input string for all potential IterationSets
        # This will produce a list of IterationSets, created from string
        iteration_strings = Tokenizer.find_all(input_string, format_string)

        # Splits the input string by the format string
        # This will produce all the strings that are not potential IterationSets
        iteration_format = Tokenizer.to_regex(format_string)
        strings = re.split(iteration_format, input_string, flags=re.I)

        iteration_sets = [
            IterationSet.from_string(x, format_string) for x in iteration_strings
        ]

        # Take both the list of IterationSets and strings and,
        # reassemble them into a single sequential list
        if strings and input_string.startswith(strings[0]):
            items = [
                x for x in chain(*zip_longest(strings, iteration_sets)) if x is not None
            ]
        else:
            items = [
                x for x in chain(*zip_longest(iteration_sets, strings)) if x is not None
            ]

        return items

    @staticmethod
    def merge(partitioned_path1, partitioned_path2, iteration_set_index=0):
        """
        Merge 2 PartitionedFilepaths together by comparing the strings and iteration sets.
        If the strings do not equal each other, a MergeError is thrown.
        IterationSets will throw a MergeError or IterationPaddingMisMatch error
         if 2 FrameSets cannot be merged.
        """

        if len(partitioned_path1) != len(partitioned_path2):
            raise MergeError("Partitioned paths are different length")

        merged = []
        iteration_set_counter = 0
        for item1, item2 in zip(partitioned_path1, partitioned_path2):
            if isinstance(item1, IterationSet) and isinstance(item2, IterationSet):

                to_append = item1
                if iteration_set_counter == iteration_set_index:
                    merged_iteration_set = IterationSet.merge(item1, item2)
                    to_append = merged_iteration_set
                elif item1 != item2:
                    raise MergeError(
                        'Cannot merge "{}" -> "{}", \n{}\n->{}'.format(
                            item1, item2, partitioned_path1, partitioned_path2
                        )
                    )

                merged.append(to_append)
                iteration_set_counter += 1

            elif item1 != item2:
                raise MergeError(
                    'Cannot merge "{}" -> "{}", \n{}\n->{}'.format(
                        item1, item2, partitioned_path1, partitioned_path2
                    )
                )
            else:
                merged.append(item1)

        new_partitioned_path = PartitionedFilepath.from_items(merged)

        return new_partitioned_path

    @staticmethod
    def compress_partitions(partitioned_filepaths, iteration_set_index=0):

        already_merged = []
        for partitioned_path in partitioned_filepaths:

            if not already_merged:
                already_merged.append(partitioned_path)
                continue

            merged = False
            for i, merged_path in enumerate(already_merged):
                try:
                    new_partition = PartitionedFilepath.merge(
                        partitioned_path,
                        merged_path,
                        iteration_set_index=iteration_set_index,
                    )
                    already_merged[i] = new_partition
                    merged = True
                    break
                except MergeError:
                    pass

            if not merged:
                already_merged.append(partitioned_path)

        return already_merged

    def iteration_sets(self):
        return list(self._iteration_sets.values())

    def strings(self):
        return self._strings.values()

    def template(self):
        ret_str = ""
        for item in self:
            if isinstance(item, IterationSet):
                ret_str += "<iteration_set>"
            else:
                ret_str += "<string>"
        return ret_str


class CompressedFilepath(object):

    TEMPLATE_STRINGS = ()
    STR_REPR_FORMAT = ""

    def __init__(self, partitioned_filepath):
        self._partitioned_filepath = partitioned_filepath
        self._size = 0

    def __str__(self):
        return str(self._partitioned_filepath)

    def __repr__(self):
        return str(self._partitioned_filepath)

    def __len__(self):
        # length = 0
        # for iteration_set in self._partitioned_filepath.iteration_sets():
        #     length += len(iteration_set)
        # return length
        return len(self.expand())

    def __iter__(self):
        for item in self.expand():
            yield item

    @classmethod
    def compress(cls, strings, include=None, exclude=None):

        strings = filter_string_list(strings, include=include, exclude=exclude)

        partitioned_filepaths = map(PartitionedFilepath, strings)

        sorted_partitions = defaultdict(list)

        for partitioned_path in partitioned_filepaths:
            sorted_partitions[len(partitioned_path.iteration_sets())].append(
                partitioned_path
            )

        total_merged = []
        for num_of_iteration_sets, partitioned_paths in sorted_partitions.items():
            for i in reversed(range(num_of_iteration_sets)):
                partitioned_paths = PartitionedFilepath.compress_partitions(
                    partitioned_paths, i
                )
            total_merged.extend(partitioned_paths)

        return [cls(partitioned_path) for partitioned_path in total_merged]

    @staticmethod
    def expand_many(compressed_filepaths):
        expanded = []

        for compressed_path in compressed_filepaths:
            expanded.extend(compressed_path.expand())

        return expanded

    @classmethod
    def parse(
        cls, input_string, iteration_set_format="{first}-{last}", check_disk=True
    ):

        partitioned_filepath = PartitionedFilepath.from_str_repr(
            input_string, iteration_set_format
        )

        items = []
        has_empty_iteration_sets = False
        if check_disk:
            for item in partitioned_filepath:
                if isinstance(item, IterationSet):
                    if item.is_empty():
                        has_empty_iteration_sets = True
                        items.append("*")
                else:
                    items.append(item)

        if has_empty_iteration_sets:
            glob_path = "".join(items)
            compressed = cls.compress(glob(glob_path))

            if not compressed:
                raise FilesNotLocal("{} cannot be found locally".format(glob_path))

            return compressed[0]

        return cls(partitioned_filepath)

    def format(self, iteration_set_format):
        return "".join(
            [item.format(iteration_set_format) for item in self._partitioned_filepath]
        )

    def expand(self):
        path_strings = []
        for item in self._partitioned_filepath:

            if not path_strings:
                if isinstance(item, IterationSet):
                    path_strings.extend(list(item.str_iter()))
                else:
                    path_strings.append(item)
                continue

            if isinstance(item, IterationSet):
                new_strings = []
                for iteration in item.str_iter():
                    new_strings.extend(
                        ["{}{}".format(x, iteration) for x in path_strings]
                    )
                path_strings = new_strings
            else:
                path_strings = ["{}{}".format(x, item) for x in path_strings]

        return path_strings

    def has(self, string):
        return string in self.expand()

    def relative(self):
        return os.path.relpath(str(self))

    def absolute(self):
        return os.path.abspath(str(self))

    def realpath(self):
        return os.path.realpath(str(self))

    def str_repr(self, iteration_format=None):

        if not self.STR_REPR_FORMAT:
            return self.format(iteration_format)

        try:
            return self.STR_REPR_FORMAT.format(**self._filepath_attributes)
        except KeyError:
            return self.format(iteration_format)

    def size(self):
        if self._size:
            return self._size

        for item in self.expand():
            try:
                self._size += os.stat(item).st_size
            except OSError:
                # This isn't assuming the filepaths are local
                pass

        return self._size


def filter_string_list(string_list, include=None, exclude=None):
    """
    Filter a list of filepaths based on include and exclude patterns
    The current behavior:
        includes:
            loop through the includes:
                each include is applied to the set of paths that is the result of the previous filter
        excludes:
            loop through the excludes:
                each exclude is applied to the set of paths that is the result of the previous filter
        if there is a conflict between the includes and the excludes, the exclude wins

    :param sequence_item_list: list of filepath strings
    :type sequence_item_list: list
    :param include: list of regexes to include in the return
    :type include: list
    :param exclude: list of regexes to exclude in the return
    :type exclude: list
    :return:
    """

    to_include = set(string_list)
    if include:
        if not isinstance(include, (list, set)):
            raise TypeError("include must be a list of strings/regexes")

        for inc in include:
            try:
                to_include = set(x for x in to_include if re.search(inc, x, re.I))
            except RegexError:
                continue

    if exclude:
        if not isinstance(exclude, (list, set)):
            raise TypeError("exclude must be a list of strings/regexes")

        for exc in exclude:
            try:
                to_include = set(x for x in to_include if not re.search(exc, x, re.I))
            except RegexError:
                continue

    return list(to_include)



from ._base import CompressedFilepath
from ..config import COMPRESSED_FILEPATH_TYPES


class CompressedFilepathFactory(object):
    @staticmethod
    def make_cls(name, class_config, parent_cls="CompressedFilepath"):

        # I want to protect against bad configs
        parent_cls = parent_cls or "CompressedFilepath"
        parent_cls = globals().get(parent_cls)
        if not parent_cls:
            parent_cls = parent_cls = globals().get("CompressedFilepath")

        class_data = {
            "TEMPLATE_STRINGS": tuple(
                class_config.get("filepath_attributes_parsing", [])
            ),
            "STR_REPR_FORMAT": class_config.get("str_repr_format", ""),
        }

        return type(name, (parent_cls,), class_data)


# Load any types from config
for _name, _data in COMPRESSED_FILEPATH_TYPES.items():

    _compressed_filepath = CompressedFilepathFactory.make_cls(
        name=_name, class_config=_data, parent_cls=_data.get("parent_class")
    )
    globals()[_name] = _compressed_filepath


# Subclassing below ##############################



import os
import json


__FILE_DIR = os.path.dirname(__file__)
__FORMAT_CONFIG = os.path.join(__FILE_DIR, "config.json")


CONFIG = {}
ITERATION_KEY = "iteration_formats"
COMPRESSED_FILEPATH_TYPES_KEY = "compressed_filepath_types"


if os.path.exists(__FORMAT_CONFIG):
    with open(__FORMAT_CONFIG) as f:
        try:
            CONFIG = json.loads(f.read())
        except:
            pass


ITERATION_FORMATTING = CONFIG.get(ITERATION_KEY, {})
COMPRESSED_FILEPATH_TYPES = CONFIG.get(COMPRESSED_FILEPATH_TYPES_KEY, {})



"""
Errors used in all modules of compressed_filepath
"""
from .tokenize import TokenParseError


class IterationPaddingMisMatch(ValueError):
    """
    If the padding of a IterationSet is set and the Iteration being added does not match
    """


class IterationSetParseError(Exception):
    """
    Unable to convert a formatted string to IterationSet
    """


class MergeError(Exception):
    """
    When objects of 2 different types can't merge or when 2 partitioned filepaths cannot merge
    """


class StringReprFormatError(KeyError):
    """
    When you run str_repr on a compressed filepath type but,
    the defined format template is unable to be fufilled by
    the tokens parsed from the filepath.
    """


class FilesNotLocal(Exception):
    """
    Unable to find files on disk to assist with created a CompressedFilepath
    """



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['vfx_fs_util', 'vfx_fs_util.compressed_filepath']

package_data = \
{'': ['*']}

install_requires = \
['Lucidity>=1.6.0,<2.0.0', 'scandir>=1.10.0,<2.0.0', 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['fsu = vfx_fs_util.__main__:main']}

setup_kwargs = {
    'name': 'vfx-fs-util',
    'version': '0.2.6',
    'description': 'Utility package for working filesystems in the VFX industry.',
    'long_description': None,
    'author': 'John Andrews',
    'author_email': 'john.andrews.drum@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os

try:
    from scandir import walk
except ImportError:
    from os import walk

from .compressed_filepath.types import *
from .compressed_filepath.iteration import IterationSet
from .compressed_filepath.tokenize import Tokenizer
from .compressed_filepath._base import filter_string_list


def expand_filepath(filepath):
    """
    Expand user home and any env variables
    """

    filepath = os.path.expanduser(filepath)
    filepath = os.path.expandvars(filepath)
    return filepath


def walk_dir(directory, recursive=True, include=None, exclude=None):
    """
    Walk directory and return compressed filepaths
    """

    join = os.path.join
    abspath = os.path.abspath
    is_dir = os.path.isdir

    directory = expand_filepath(directory)

    if recursive:
        for relative_dir, _, files in walk(directory):
            filepaths = [abspath(join(relative_dir, filename)) for filename in files]
            for compressed in compress_filepaths(
                filepaths, include=include, exclude=exclude
            ):
                yield compressed
    else:
        filepaths = []
        for item in os.listdir(directory):
            item_path = abspath(join(directory, item))
            if not is_dir(item_path):
                filepaths.append(item_path)

        for compressed in compress_filepaths(
            filepaths, include=include, exclude=exclude
        ):
            yield compressed


def compress_filepaths(filepaths, include=None, exclude=None, cls=CompressedFilepath):
    return cls.compress(filepaths, include=include, exclude=exclude)


def expand_compressions(compressed_filepaths, cls=CompressedFilepath):
    return cls.expand_many(compressed_filepaths)


def get_human_readable(size, base=1024):
    """
    Give it bytes and it will convert that to human readable
    ** mainly for displaying to user **
    :param size: int(bytes)
    :return: str(human readable size)
    """

    if not isinstance(size, (int, float)):
        raise AssertionError(
            "fs_util.get_human_readable function requires the size argument to be an int or float"
        )

    size_types = ["B", "KB", "MB", "GB", "TB"]

    for i, _ in enumerate(sorted(size_types, reverse=True)):
        i = len(size_types) - i
        if size > base**i:
            return "{0:.2f} {1}".format(size / float(base) ** i, size_types[i])

    return "{0:.2f} {1}".format(float(size), "B")


def mkdirs(path, mode, owner=-1, group=-1):
    assert isinstance(owner, int), "the owner must be an int representing a uid"
    assert isinstance(group, int), "the group must be an int representing a gid"
    s = path.split(os.sep)
    no_exists = []
    while not os.path.exists(path):
        no_exists.insert(0, s.pop())
        path = os.sep.join(s)
    new_path = path
    for x in no_exists:
        new_path = os.path.join(new_path, x)
        try:
            os.mkdir(new_path)
        except OSError as e:
            if "errno 17" in str(e).lower():
                continue
            else:
                raise
        if isinstance(mode, int):
            os.chmod(new_path, mode)
        os.chown(new_path, owner, group)


def find_shared_folder(file_paths):
    """
    Given a list of filepaths, find the common directory
     --> (os.path.commonprefix will consider the whole string including filename)
    :param file_paths: list(str(filepath)s)
    :return: str(filepath)
    """
    file_paths = sorted(file_paths, key=lambda path: path.count("/"))

    dir_path = os.path.dirname(file_paths[0])
    while True:
        if all(p.startswith(dir_path) for p in file_paths):
            break
        dir_path = os.path.dirname(dir_path)

    return dir_path


def from_string(path_string, check_disk=True):
    """
    From a path string, decide the function to use and return CompressedFilepaths
    """
    return CompressedFilepath.from_string(path_string, check_disk=check_disk)



import os
import sys
from typing import List

from typer import Typer, Option, Argument
from rich.console import Console
from rich.tree import Tree

from vfx_fs_util.api import (
    walk_dir,
    get_human_readable,
    compress_filepaths,
)

app = Typer()
console = Console()
err_console = Console(stderr=True, style="bold red")


@app.command()
def ls(
    directories: List[str],
    include: str = Option(
        "", "-i", "--include", help="commma separated list of regexes to include"
    ),
    exclude: str = Option(
        "", "-e", "--exclude", help="commma separated list of regexes to exclude"
    ),
    size: bool = Option(False, "-s", "--size", help="size in bytes"),
    recursive: bool = Option(
        False, "-r", "--recursive", help="Recurse through directories"
    ),
    human_readable: bool = Option(
        False, "-m", "--human_readable", help="human readable size"
    ),
    format: str = Option(None, "-f", "--format", help="choose iteration set format"),
    summary: bool = Option(
        False, "--summary", help="Output summary of files and total size"
    ),
):

    summary_map = {
        "files": 0,
        "size": 0,
    }

    include = include.split(",") if include else None
    exclude = exclude.split(",") if exclude else None

    for directory in directories:
        for compressed_path in walk_dir(
            directory, recursive=recursive, include=include, exclude=exclude
        ):
            output = compressed_path.format(format)

            if size:
                size = compressed_path.size()
                if summary:
                    summary_map["size"] += size
                if human_readable:
                    size = get_human_readable(size)
                output = "{:<15}{}".format(size, output)

            if summary:
                summary_map["files"] += len(compressed_path)

            console.print(output)

    if summary:
        if human_readable:
            summary_map["size"] = get_human_readable(summary_map.get("size"))
        console.print("files: {files}, size: {size}".format(**summary_map))


@app.command()
def tree(
    directory: str = Argument(os.getcwd()),
    include: List[str] = Option(
        [], "-i", "--include", help="list of regexes to include"
    ),
    exclude: List[str] = Option(
        [], "-e", "--exclude", help="list of regexes to exclude"
    ),
    # size: bool = Option(False, "-s", "--size", help="size in bytes"),
    level: int = Option(
        -1, "-l", "--level", help="How many levels to recurse into a directory."
    ),
    # human_readable: bool = Option(
    #     False, "-m", "--human_readable", help="human readable size"
    # ),
    # format: str = Option(None, "-f", "--format", help="choose iteration set format"),
    # summary: bool = Option(
    #     False, "--summary", help="Output summary of files and total size"
    # ),
):   
    tree = Tree(directory, guide_style="magenta")
    build_tree(
        tree,
        directory,
        include=include,
        exclude=exclude,
        level=0,
        target_level=level,
        format=format,
        # size=size,
        # human_readable=human_readable,
        # summary=summary,
    )
    console.print(tree)


def build_tree(
    tree, 
    directory, 
    level, 
    target_level,
    include=None, 
    exclude=None, 
    format=None, 
    # human_readable=False, 
    # size=False,
    # summary=False,
):

    if level==target_level:
        return

    files = []
    for item in os.listdir(directory):
        filepath = os.path.join(directory, item)
        if os.path.isdir(filepath):
            subtree = tree.add(item)
            level+=1
            build_tree(
                subtree,
                filepath, 
                level=level, 
                target_level=target_level,
                include=include, 
                exclude=exclude
            )
        else:
            files.append(item)

    compressed_paths = compress_filepaths(files, include=include, exclude=exclude)
    for compressed_file in compressed_paths:
        filepath_string = compressed_file.format(format)
        tree.add(filepath_string)


def main():
    try:
        app()
    except Exception as err:
        err_console.print(err)
        sys.exit(1)


if __name__ == "__main__":
    sys.exit(main())
